import enum
import typing as t

import cattr

from .events import (
    Initialized,
    Completion,
    ServerRequest,
    Shutdown,
    PublishDiagnostics,
    Event,
    ShowMessage,
    ServerNotification,
    WillSaveWaitUntilEdits,
    ShowMessageRequest,
    LogMessage,
)
from .structs import (
    Response,
    TextDocumentPosition,
    CompletionContext,
    CompletionList,
    CompletionItem,
    Request,
    JSONDict,
    MessageActionItem,
    MessageType,
    TextDocumentItem,
    TextDocumentIdentifier,
    VersionedTextDocumentIdentifier,
    TextDocumentContentChangeEvent,
    TextDocumentSaveReason,
    TextEdit,
)
from .io_handler import _make_request, _parse_messages, _make_response


class ClientState(enum.Enum):
    NOT_INITIALIZED = enum.auto()
    WAITING_FOR_INITIALIZED = enum.auto()
    NORMAL = enum.auto()
    WAITING_FOR_SHUTDOWN = enum.auto()
    SHUTDOWN = enum.auto()
    EXITED = enum.auto()


class Client:
    # TODO: Save the encoding given here.
    def __init__(
        self,
        process_id: t.Optional[int] = None,
        root_uri: t.Optional[str] = None,
        trace: str = "off",
    ) -> None:
        self._state = ClientState.NOT_INITIALIZED

        # Used to save data as it comes in (from `recieve_bytes`) until we have
        # a full request.
        self._recv_buf = bytearray()

        # Things that we still need to send.
        self._send_buf = bytearray()

        # Keeps track of which IDs match to which unanswered requests.
        self._unanswered_requests: t.Dict[int, Request] = {}

        # Just a simple counter to make sure we have unique IDs. We could make
        # sure that this fits into a JSONRPC Number, seeing as Python supports
        # bignums, but I think that's an unlikely enough case that checking for
        # it would just litter the code unnecessarily.
        self._id_counter = 0

        # We'll just immediately send off an "initialize" request.
        self._send_request(
            method="initialize",
            params={
                "processId": process_id,
                "rootUri": root_uri,
                "capabilities": {},
                "trace": trace,
            },
        )
        self._state = ClientState.WAITING_FOR_INITIALIZED

    @property
    def state(self) -> ClientState:
        return self._state

    def _send_request(
        self, method: str, params: t.Optional[JSONDict] = None
    ) -> int:
        id = self._id_counter
        self._id_counter += 1

        self._send_buf += _make_request(method=method, params=params, id=id)
        self._unanswered_requests[id] = Request(
            id=id, method=method, params=params
        )
        return id

    def _send_notification(
        self, method: str, params: t.Optional[JSONDict] = None
    ) -> None:
        self._send_buf += _make_request(method=method, params=params)

    def _send_response(
        self,
        id: int,
        result: t.Optional[JSONDict] = None,
        error: t.Optional[JSONDict] = None,
    ) -> None:
        self._send_buf += _make_response(id=id, result=result, error=error)

    def recv(self, data: bytes) -> t.List[Event]:
        self._recv_buf += data

        # _parse_messages deletes consumed data from self._recv_buf
        messages = list(_parse_messages(self._recv_buf))

        events = []
        for message in messages:
            if isinstance(message, Response):
                response = message
                assert response.id is not None
                request = self._unanswered_requests.pop(response.id)

                # FIXME: The errors have meanings.
                if response.error is not None:
                    __import__("pprint").pprint(response.error)
                    raise RuntimeError("Response error!")

                if request.method == "initialize":
                    assert self._state == ClientState.WAITING_FOR_INITIALIZED
                    self._send_notification("initialized")
                    events.append(
                        cattr.structure(response.result, Initialized)
                    )
                    self._state = ClientState.NORMAL
                elif request.method == "shutdown":
                    assert self._state == ClientState.WAITING_FOR_SHUTDOWN
                    events.append(Shutdown())
                    self._state = ClientState.SHUTDOWN
                elif request.method == "textDocument/completion":
                    completion_list = None

                    try:
                        completion_list = cattr.structure(
                            response.result, CompletionList
                        )
                    except TypeError:
                        try:
                            completion_list = CompletionList(
                                isIncomplete=False,
                                items=cattr.structure(
                                    response.result, t.List[CompletionItem]
                                ),
                            )
                        except TypeError:
                            assert response.result is None

                    events.append(Completion(response.id, completion_list))
                elif request.method == "textDocument/willSaveWaitUntil":
                    # FIXME: types don't match???
                    events.append(
                        WillSaveWaitUntilEdits(
                            edits=t.cast(t.List[TextEdit], response.result)
                        )
                    )
                else:
                    raise NotImplementedError((response, request))
            elif isinstance(message, Request):
                request = message

                E = t.TypeVar("E", bound=Event)

                def structure_request(event_cls: t.Type[E]) -> E:
                    if issubclass(event_cls, ServerRequest):
                        event = cattr.structure(request.params, event_cls)
                        assert request.id is not None
                        event._id = request.id
                        event._client = self
                        return t.cast("E", event)
                    elif issubclass(event_cls, ServerNotification):
                        return t.cast(
                            "E", cattr.structure(request.params, event_cls)
                        )
                    else:
                        raise TypeError(
                            "`event_cls` must be a subclass of ServerRequest or ServerNotification"
                        )

                if request.method == "window/showMessage":
                    events.append(structure_request(ShowMessage))
                elif request.method == "window/showMessageRequest":
                    events.append(structure_request(ShowMessageRequest))
                elif request.method == "window/logMessage":
                    events.append(structure_request(LogMessage))
                elif request.method == "textDocument/publishDiagnostics":
                    events.append(structure_request(PublishDiagnostics))
                else:
                    raise NotImplementedError(request)
            else:
                raise RuntimeError("nobody will ever see this, i hope")

        return events

    def send(self) -> bytes:
        send_buf = self._send_buf[:]
        self._send_buf.clear()
        return send_buf

    def shutdown(self) -> None:
        assert self._state == ClientState.NORMAL
        self._send_request(method="shutdown")
        self._state = ClientState.WAITING_FOR_SHUTDOWN

    def exit(self) -> None:
        assert self._state == ClientState.SHUTDOWN
        self._send_notification(method="exit")
        self._state = ClientState.EXITED

    def cancel_last_request(self) -> None:
        self._send_notification(
            method="$/cancelRequest", params={"id": self._id_counter - 1}
        )

    def did_open(self, text_document: TextDocumentItem) -> None:
        assert self._state == ClientState.NORMAL
        self._send_notification(
            method="textDocument/didOpen",
            params={"textDocument": cattr.unstructure(text_document)},
        )

    def did_change(
        self,
        text_document: VersionedTextDocumentIdentifier,
        content_changes: t.List[TextDocumentContentChangeEvent],
    ) -> None:
        assert self._state == ClientState.NORMAL

        self._send_notification(
            method="textDocument/didChange",
            params={
                "textDocument": cattr.unstructure(text_document),
                "contentChanges": cattr.unstructure(content_changes),
            },
        )

    def will_save(
        self,
        text_document: TextDocumentIdentifier,
        reason: TextDocumentSaveReason,
    ) -> None:
        assert self._state == ClientState.NORMAL
        self._send_notification(
            method="textDocument/willSave",
            params={
                "textDocument": cattr.unstructure(text_document),
                "reason": cattr.unstructure(reason),
            },
        )

    def will_save_wait_until(
        self,
        text_document: TextDocumentIdentifier,
        reason: TextDocumentSaveReason,
    ) -> None:
        assert self._state == ClientState.NORMAL
        self._send_request(
            method="textDocument/willSaveWaitUntil",
            params={
                "textDocument": cattr.unstructure(text_document),
                "reason": cattr.unstructure(reason),
            },
        )

    def did_save(
        self,
        text_document: TextDocumentIdentifier,
        text: t.Optional[str] = None,
    ) -> None:
        assert self._state == ClientState.NORMAL
        params = {"textDocument": cattr.unstructure(text_document)}
        if text is not None:
            params["text"] = text
        self._send_notification(method="textDocument/didSave", params=params)

    def did_close(self, text_document: TextDocumentIdentifier) -> None:
        assert self._state == ClientState.NORMAL
        self._send_notification(
            method="textDocument/didClose",
            params={"textDocument": cattr.unstructure(text_document)},
        )

    def completions(
        self,
        text_document_position: TextDocumentPosition,
        context: t.Optional[CompletionContext] = None,
    ) -> int:
        assert self._state == ClientState.NORMAL
        params = {}
        params.update(cattr.unstructure(text_document_position))
        if context is not None:
            params.update(cattr.unstructure(context))
        return self._send_request(
            method="textDocument/completion", params=params
        )



import cgi
import json
import typing as t

import cattr

from .structs import Request, Response, JSONDict


def _make_headers(content_length: int, encoding: str = "utf-8") -> bytes:
    headers_bytes = bytearray()
    headers = {
        "Content-Length": content_length,
        "Content-Type": f"application/vscode-jsonrpc; charset={encoding}",
    }
    for (key, value) in headers.items():
        headers_bytes += f"{key}: {value}\r\n".encode(encoding)
    headers_bytes += b"\r\n"
    return headers_bytes


def _make_request(
    method: str,
    params: t.Optional[JSONDict] = None,
    id: t.Optional[int] = None,
    *,
    encoding: str = "utf-8",
) -> bytes:
    request = bytearray()

    # Set up the actual JSONRPC content and encode it.
    content: JSONDict = {"jsonrpc": "2.0", "method": method}
    if params is not None:
        content["params"] = params
    if id is not None:
        content["id"] = id
    encoded_content = json.dumps(content).encode(encoding)

    # Write the headers to the request body
    request += _make_headers(
        content_length=len(encoded_content), encoding=encoding
    )

    # Append the content to the request
    request += encoded_content

    return request


def _make_response(
    id: int,
    result: t.Optional[JSONDict] = None,
    error: t.Optional[JSONDict] = None,
    *,
    encoding: str = "utf-8",
) -> bytes:
    request = bytearray()

    # Set up the actual JSONRPC content and encode it.
    content: JSONDict = {"jsonrpc": "2.0", "id": id}
    if result is not None:
        content["result"] = result
    if error is not None:
        content["error"] = error
    encoded_content = json.dumps(content).encode(encoding)

    # Write the headers to the request body
    request += _make_headers(
        content_length=len(encoded_content), encoding=encoding
    )

    # Append the content to the request
    request += encoded_content

    return request


# _parse_messages is kind of tricky.
#
# It used to work like this:
#   - Take in bytes argument
#   - Yield parsed things (Responses and Requests)
#   - Raise IncompleteResponseError if parsing fails by running out of bytes
#
# Then, if the bytes contained two valid things (Responses or Requests) and
# then only a part of another thing, this function would first parse and yield
# the valid things and then raise IncompleteResponseError, outputting no
# information about which bytes hadn't been parsed yet. The code calling this
# function would see only the IncompleteResponseError and discard the parsed
# things. Even with that bugginess, this actually worked because every once in
# a while there would be no bytes left from parsing things, and this function
# would output parsed things with no error.
#
# The least error-prone fix I could think of is to pass the bytes to parse
# as a mutable bytearray object, and just delete the used bytes from there.
# This means that we no longer have any need for IncompleteResponseError, and
# this function will return an empty list instead in that case.
#
# _parse_one_message returns None when there are no more messages, and an empty
# iterator when a message was parsed but no things were created.
def _parse_one_message(
    response_buf: bytearray,
) -> t.Optional[t.Iterable[t.Union[Response, Request]]]:
    if b"\r\n\r\n" not in response_buf:
        return None

    header_lines, raw_content = bytes(response_buf).split(b"\r\n\r\n", 1)

    # Many langservers don't set Content-Type header for whatever reason. We
    # use a sane default for that.
    #
    # Langserver spec links to RFC 7230 which says that header names should be
    # case-insensitive.
    headers = {"content-type": "application/vscode-jsonrpc; charset=utf-8"}
    for header_line in header_lines.split(b"\r\n"):
        key, value = header_line.decode("ascii").split(": ", 1)
        headers[key.lower()] = value

    # We will now parse the Content-Type and Content-Length headers. Since for
    # version 3.0 of the Language Server Protocol they're the only ones, we can
    # just verify they're there and not keep them around in the Response
    # object.
    assert set(headers.keys()) == {"content-type", "content-length"}

    # Content-Type and encoding.
    content_type, metadata = cgi.parse_header(headers["content-type"])
    assert content_type == "application/vscode-jsonrpc"
    encoding = metadata["charset"]

    # Content-Length
    content_length = int(headers["content-length"])

    # We need to verify that the raw_content is long enough.
    if len(raw_content) < content_length:
        # incomplete request
        return None

    # Take only as many bytes as we need. If there's any remaining, they're
    # the next response's.
    raw_content = raw_content[:content_length]
    unused_bytes_count = len(raw_content[content_length:])

    # This is a good place for deleting unnecessary stuff from response_buf
    # because if the code below fails, then leaving the cause of failure to
    # response_buf would cause this function to fail every time in the future
    # when called with the same response_buf. I think I've had this issue a
    # long time ago, and it was annoying how one response parsing error would
    # also block the parsing of any future responses.
    if unused_bytes_count == 0:  # 'del response_buf[:-0]' does wrong thing
        response_buf.clear()
    else:
        del response_buf[:-unused_bytes_count]

    def do_it(data: JSONDict) -> t.Union[Response, Request]:
        del data["jsonrpc"]

        # Request must come first because it has a non-optional attrib
        #
        # that is, any Request is also a valid Response (everything just gets
        # filled in with None)
        try:
            request = cattr.structure(data, Request)
            assert isinstance(request, Request)
            return request
        except TypeError:
            pass

        try:
            response = cattr.structure(data, Response)
            assert isinstance(response, Response)
            return response
        except TypeError:
            pass

        raise RuntimeError(f"{data!r} is neither a Request nor a Response!")

    content = json.loads(raw_content.decode(encoding))
    if isinstance(content, list):
        # This is in response to a batch operation.
        return map(do_it, content)
    else:
        return [do_it(content)]


def _parse_messages(
    response_buf: bytearray,
) -> t.Iterator[t.Union[Response, Request]]:
    while True:
        parsed = _parse_one_message(response_buf)
        if parsed is None:
            break
        yield from parsed



import typing as t

import attr
from attr import attrs, attrib

if t.TYPE_CHECKING:  # avoid import cycle at runtime
    from .client import Client
from .structs import (
    JSONDict,
    Diagnostic,
    Range,
    MessageType,
    MessageActionItem,
    CompletionItem,
    CompletionList,
    TextEdit,
)


@attrs
class Event:
    pass


@attrs
class ServerRequest(Event):
    _client: "Client" = attrib(init=False)
    _id: int = attrib(init=False)


@attrs
class ServerNotification(Event):
    pass


@attrs
class Initialized(Event):
    capabilities: JSONDict = attrib()


@attrs
class Shutdown(Event):
    pass


@attrs
class ShowMessage(ServerNotification):
    type: MessageType = attrib()
    message: str = attrib()


@attrs
class ShowMessageRequest(ServerRequest):
    type: MessageType = attrib()
    message: str = attrib()
    actions: t.Optional[t.List[MessageActionItem]] = attrib()

    def reply(self, action: t.Optional[MessageActionItem] = None) -> None:
        """
        Reply to the ShowMessageRequest with the user's selection.

        No bytes are actually returned from this method, the reply's bytes are added to the client's internal send buffer.
        """
        self._client._send_response(id=self._id, result=attr.asdict(action))


@attrs
class LogMessage(ServerNotification):
    type: MessageType = attrib()
    message: str = attrib()


@attrs
class Completion:
    message_id: int = attrib()
    completion_list: t.Union[
        CompletionList, t.List[CompletionItem], None
    ] = attrib()


# XXX: not sure how to name this event.
@attrs
class WillSaveWaitUntilEdits:
    edits: t.Optional[t.List[TextEdit]] = attrib(default=None)


@attrs
class PublishDiagnostics(ServerNotification):
    uri: str = attrib()
    diagnostics: t.List[Diagnostic] = attrib()



import enum
import typing as t

from attr import attrs, attrib

# XXX: Replace the non-commented-out code with what's commented out once nested
# types become a thing in mypy.
# JSONValue = t.Union[None, str, int, t.List['JSONValue'], t.Dict[str, 'JSONValue']]
# JSONDict = t.Dict[str, JSONValue]
JSONDict = t.Dict[str, t.Any]

# XXX: We can't have this be both str and int due to `cattrs` restrictions. How
# can we fix this?
# Id = t.Union[str, int]
Id = int


@attrs
class Request:
    method: str = attrib()
    id: t.Optional[Id] = attrib(default=None)
    params: t.Optional[JSONDict] = attrib(default=None)


@attrs
class Response:
    id: t.Optional[Id] = attrib(default=None)
    result: t.Optional[JSONDict] = attrib(default=None)
    error: t.Optional[JSONDict] = attrib(default=None)


class MessageType(enum.IntEnum):
    ERROR = 1
    WARNING = 2
    INFO = 3
    LOG = 4


@attrs
class MessageActionItem:
    title: str = attrib()


@attrs
class TextDocumentItem:
    uri: str = attrib()
    languageId: str = attrib()
    version: int = attrib()
    text: str = attrib()


@attrs
class TextDocumentIdentifier:
    uri: str = attrib()


@attrs
class VersionedTextDocumentIdentifier(TextDocumentIdentifier):
    version: t.Optional[int] = attrib(default=None)


@attrs
class Position:
    # NB: These are both zero-based.
    line: int = attrib()
    character: int = attrib()


@attrs
class Range:
    start: Position = attrib()
    end: Position = attrib()

    def calculate_length(self, text: str) -> int:
        text_lines = text.splitlines()

        if self.end.line == self.start.line:
            line = text_lines[self.start.line]
            return len(line[self.start.character : self.end.character])
        else:
            total = 0

            total += len(text_lines[self.start.line][self.start.character :])

            for line_number in range(self.start.line + 1, self.end.line):
                total += len(text_lines[line_number])

            total += len(text_lines[self.end.line][: self.end.character])

            return total


@attrs
class TextDocumentContentChangeEvent:
    text: str = attrib()
    range: t.Optional[Range] = attrib(default=None)
    rangeLength: t.Optional[int] = attrib(default=None)

    @classmethod
    def change_range(
        cls,
        change_start: Position,
        change_end: Position,
        change_text: str,
        old_text: str,
    ) -> "TextDocumentContentChangeEvent":
        """
        Create a TextDocumentContentChangeEvent reflecting the given changes.

        Nota bene: If you're creating a list of TextDocumentContentChangeEvent based on many changes,
        `old_text` must reflect the state of the text after all previous change events happened.
        Or you can just use `sansio_lsp_client.utils.calculate_change_events`.
        """
        change_range = Range(change_start, change_end)
        return cls(
            range=change_range,
            rangeLength=change_range.calculate_length(old_text),
            text=change_text,
        )

    @classmethod
    def change_whole_document(
        cls, change_text: str
    ) -> "TextDocumentContentChangeEvent":
        return cls(text=change_text)


@attrs
class TextDocumentPosition:
    textDocument: TextDocumentIdentifier = attrib()
    position: Position = attrib()


class CompletionTriggerKind(enum.IntEnum):
    INVOKED = 1
    TRIGGER_CHARACTER = 2
    TRIGGER_FOR_INCOMPLETE_COMPLETIONS = 3


@attrs
class CompletionContext:
    triggerKind: CompletionTriggerKind = attrib()
    triggerCharacter: t.Optional[str] = attrib(default=None)


class MarkupKind(enum.Enum):
    PLAINTEXT = "plaintext"
    MARKDOWN = "markdown"


@attrs
class MarkupContent:
    kind: MarkupKind = attrib()
    value: str = attrib()


@attrs
class TextEdit:
    range: Range = attrib()
    newText: str = attrib()


@attrs
class Command:
    title: str = attrib()
    command: str = attrib()
    arguments: t.Optional[t.List[t.Any]] = attrib(default=None)


class InsertTextFormat(enum.IntEnum):
    PLAIN_TEXT = 1
    SNIPPET = 2


@attrs
class CompletionItem:
    label: str = attrib()
    # TODO: implement CompletionItemKind.
    kind: t.Optional[int] = attrib(default=None)
    detail: t.Optional[str] = attrib(default=None)
    # FIXME: Allow `t.Union[str, MarkupContent]` here by defining a cattrs
    # custom loads hook.
    documentation: t.Optional[str] = attrib(default=None)
    deprecated: t.Optional[bool] = attrib(default=None)
    preselect: t.Optional[bool] = attrib(default=None)
    sortText: t.Optional[str] = attrib(default=None)
    filterText: t.Optional[str] = attrib(default=None)
    insertText: t.Optional[str] = attrib(default=None)
    insertTextFormat: t.Optional[InsertTextFormat] = attrib(default=None)
    textEdit: t.Optional[TextEdit] = attrib(default=None)
    additionalTextEdits: t.Optional[t.List[TextEdit]] = attrib(default=None)
    commitCharacters: t.Optional[t.List[str]] = attrib(default=None)
    command: t.Optional[Command] = attrib(default=None)
    data: t.Optional[t.Any] = attrib(default=None)


@attrs
class CompletionList:
    isIncomplete: bool = attrib()
    items: t.List[CompletionItem] = attrib()


class TextDocumentSaveReason(enum.IntEnum):
    MANUAL = 1
    AFTER_DELAY = 2
    FOCUS_OUT = 3


@attrs
class Location:
    uri: str = attrib()
    range: Range = attrib()


@attrs
class DiagnosticRelatedInformation:
    location: Location = attrib()
    message: str = attrib()


class DiagnosticSeverity(enum.IntEnum):
    ERROR = 1
    WARNING = 2
    INFORMATION = 3
    HINT = 4


@attrs
class Diagnostic:
    range: Range = attrib()

    # TODO: Make this a proper enum
    severity: DiagnosticSeverity = attrib(default=None)

    # TODO: Support this as an union of str and int
    code: t.Optional[t.Any] = attrib(default=None)

    source: t.Optional[str] = attrib(default=None)

    message: t.Optional[str] = attrib(default=None)

    relatedInformation: t.Optional[
        t.List[DiagnosticRelatedInformation]
    ] = attrib(default=None)



import typing as t
import difflib

from .structs import TextDocumentContentChangeEvent, Position

# Thanks to Akuli for writing most of this code!
# XXX: This code has been through many iterations and even now I'm not sure
# it's correct. It's pretty hard code to test, but I'll eventually get around
# to it.
def calculate_change_events(
    old_text: str, new_text: str
) -> t.List[TextDocumentContentChangeEvent]:
    seq_matcher = difflib.SequenceMatcher(a=old_text, b=new_text)
    events = []

    adjusted_text = old_text[:]

    index_offset = 0

    def index_to_position(index: int) -> Position:
        line = adjusted_text.count("\n", 0, index)
        character = index - (adjusted_text.rfind("\n", 0, index) + 1)
        return Position(line=line, character=character)

    for (
        opcode,
        old_start,
        old_end,
        new_start,
        new_end,
    ) in seq_matcher.get_opcodes():
        if opcode == "equal":
            continue

        adjusted_start_index = old_start + index_offset
        adjusted_end_index = old_end + index_offset

        # To tell how many characters we're about to insert, we just need to
        # know the length of the replacement text. However, we may be replacing
        # some characters that are already there, so we just need to subtract
        # the length of the replacement bit. This will be equal to 0 if we're
        # inserting new text.
        index_offset += len(new_text[new_start:new_end]) - (
            adjusted_end_index - adjusted_start_index
        )

        adjusted_start_pos = index_to_position(adjusted_start_index)
        adjusted_end_pos = index_to_position(adjusted_end_index)

        events.append(
            TextDocumentContentChangeEvent.change_range(
                change_start=adjusted_start_pos,
                change_end=adjusted_end_pos,
                change_text=new_text[new_start:new_end],
                old_text=adjusted_text,
            )
        )

        # Now we'll make the adjusted text actually reflect the changes.
        adjusted_text = (
            adjusted_text[:adjusted_start_index]
            + new_text[new_start:new_end]
            + adjusted_text[adjusted_end_index:]
        )
    return events



"""An implementation of the client side of the LSP protocol, useful for embedding easily in your editor."""

from .client import *
from .events import *
from .structs import *
from .utils import *

__version__ = "0.7.2"
